@@ -9,7 +9,6 @@ use warnings;
 use strict;
 use IkiWiki 3.00;
 use Encode;
-use POSIX qw(strftime);
 
 use constant PREVIEW => "Preview";
 use constant POST_COMMENT => "Post comment";
@@ -21,7 +20,8 @@ my %commentstate;
 sub import {
 	hook(type => "checkconfig", id => 'comments',  call => \&checkconfig);
 	hook(type => "getsetup", id => 'comments',  call => \&getsetup);
-	hook(type => "preprocess", id => 'comment', call => \&preprocess);
+	hook(type => "preprocess", id => 'comment', call => \&preprocess,
+		scan => 1);
 	hook(type => "preprocess", id => 'commentmoderation', call => \&preprocess_moderation);
 	# here for backwards compatability with old comments
 	hook(type => "preprocess", id => '_comment', call => \&preprocess);
@@ -35,6 +35,7 @@ sub import {
 	# Load goto to fix up user page links for logged-in commenters
 	IkiWiki::loadplugin("goto");
 	IkiWiki::loadplugin("inline");
+	IkiWiki::loadplugin("transient");
 }
 
 sub getsetup () {
@@ -90,17 +91,31 @@ sub getsetup () {
 			safe => 0,
 			rebuild => 0,
 		},
+		comments_allowformats => {
+			type => 'string',
+			default => '',
+			example => 'mdwn txt',
+			description => 'Restrict formats for comments to (no restriction if empty)',
+			safe => 1,
+			rebuild => 0,
+		},
+
 }
 
 sub checkconfig () {
 	$config{comments_commit} = 1
 		unless defined $config{comments_commit};
+	if (! $config{comments_commit}) {
+		$config{only_committed_changes}=0;
+	}
 	$config{comments_pagespec} = ''
 		unless defined $config{comments_pagespec};
 	$config{comments_closed_pagespec} = ''
 		unless defined $config{comments_closed_pagespec};
 	$config{comments_pagename} = 'comment_'
 		unless defined $config{comments_pagename};
+	$config{comments_allowformats} = ''
+		unless defined $config{comments_allowformats};
 }
 
 sub htmlize {
@@ -128,12 +143,18 @@ sub safeurl ($) {
 	}
 }
 
+sub isallowed ($) {
+    my $format = shift;
+    return ! $config{comments_allowformats} || $config{comments_allowformats} =~ /\b$format\b/;
+}
+
 sub preprocess {
 	my %params = @_;
 	my $page = $params{page};
 
 	my $format = $params{format};
-	if (defined $format && ! exists $IkiWiki::hooks{htmlize}{$format}) {
+	if (defined $format && (! exists $IkiWiki::hooks{htmlize}{$format} ||
+				! isallowed($format))) {
 		error(sprintf(gettext("unsupported page format %s"), $format));
 	}
 
@@ -143,22 +164,27 @@ sub preprocess {
 	}
 	$content =~ s/\\"/"/g;
 
-	if ($config{comments_allowdirectives}) {
-		$content = IkiWiki::preprocess($page, $params{destpage},
-			$content);
+	if (defined wantarray) {
+		if ($config{comments_allowdirectives}) {
+			$content = IkiWiki::preprocess($page, $params{destpage},
+				$content);
+		}
+
+		# no need to bother with htmlize if it's just HTML
+		$content = IkiWiki::htmlize($page, $params{destpage}, $format, $content)
+			if defined $format;
+
+		IkiWiki::run_hooks(sanitize => sub {
+			$content = shift->(
+				page => $page,
+				destpage => $params{destpage},
+				content => $content,
+			);
+		});
+	}
+	else {
+		IkiWiki::preprocess($page, $params{destpage}, $content, 1);
 	}
-
-	# no need to bother with htmlize if it's just HTML
-	$content = IkiWiki::htmlize($page, $params{destpage}, $format, $content)
-		if defined $format;
-
-	IkiWiki::run_hooks(sanitize => sub {
-		$content = shift->(
-			page => $page,
-			destpage => $params{destpage},
-			content => $content,
-		);
-	});
 
 	# set metadata, possibly overriding [[!meta]] directives from the
 	# comment itself
@@ -180,10 +206,12 @@ sub preprocess {
 			$commentopenid = $commentuser;
 		}
 		else {
-			$commentauthorurl = IkiWiki::cgiurl(
-				do => 'goto',
-				page => IkiWiki::userpage($commentuser)
-			);
+			if (length $config{cgiurl}) {
+				$commentauthorurl = IkiWiki::cgiurl(
+					do => 'goto',
+					page => IkiWiki::userpage($commentuser)
+				);
+			}
 
 			$commentauthor = $commentuser;
 		}
@@ -195,38 +223,35 @@ sub preprocess {
 		$commentauthor = gettext("Anonymous");
 	}
 
-	$commentstate{$page}{commentuser} = $commentuser;
-	$commentstate{$page}{commentopenid} = $commentopenid;
-	$commentstate{$page}{commentip} = $commentip;
-	$commentstate{$page}{commentauthor} = $commentauthor;
-	$commentstate{$page}{commentauthorurl} = $commentauthorurl;
-	if (! defined $pagestate{$page}{meta}{author}) {
-		$pagestate{$page}{meta}{author} = $commentauthor;
-	}
-	if (! defined $pagestate{$page}{meta}{authorurl}) {
-		$pagestate{$page}{meta}{authorurl} = $commentauthorurl;
-	}
-
 	if ($config{comments_allowauthor}) {
 		if (defined $params{claimedauthor}) {
-			$pagestate{$page}{meta}{author} = $params{claimedauthor};
+			$commentauthor = $params{claimedauthor};
 		}
 
 		if (defined $params{url}) {
 			my $url=$params{url};
 
 			eval q{use URI::Heuristic}; 
-		  	if (! $@) {
+			if (! $@) {
 				$url=URI::Heuristic::uf_uristr($url);
 			}
 
 			if (safeurl($url)) {
-				$pagestate{$page}{meta}{authorurl} = $url;
+				$commentauthorurl = $url;
 			}
 		}
 	}
-	else {
+
+	$commentstate{$page}{commentuser} = $commentuser;
+	$commentstate{$page}{commentopenid} = $commentopenid;
+	$commentstate{$page}{commentip} = $commentip;
+	$commentstate{$page}{commentauthor} = $commentauthor;
+	$commentstate{$page}{commentauthorurl} = $commentauthorurl;
+	$commentstate{$page}{commentauthoravatar} = $params{avatar};
+	if (! defined $pagestate{$page}{meta}{author}) {
 		$pagestate{$page}{meta}{author} = $commentauthor;
+	}
+	if (! defined $pagestate{$page}{meta}{authorurl}) {
 		$pagestate{$page}{meta}{authorurl} = $commentauthorurl;
 	}
 
@@ -295,7 +320,8 @@ sub editcomment ($$) {
 
 	my @buttons = (POST_COMMENT, PREVIEW, CANCEL);
 	my $form = CGI::FormBuilder->new(
-		fields => [qw{do sid page subject editcontent type author url}],
+		fields => [qw{do sid page subject editcontent type author
+			email url subscribe anonsubscribe}],
 		charset => 'utf-8',
 		method => 'POST',
 		required => [qw{editcontent}],
@@ -325,7 +351,7 @@ sub editcomment ($$) {
 
 	my @page_types;
 	if (exists $IkiWiki::hooks{htmlize}) {
-		foreach my $key (grep { !/^_/ } keys %{$IkiWiki::hooks{htmlize}}) {
+		foreach my $key (grep { !/^_/ && isallowed($_) } keys %{$IkiWiki::hooks{htmlize}}) {
 			push @page_types, [$key, $IkiWiki::hooks{htmlize}{$key}{longname} || $key];
 		}
 	}
@@ -340,18 +366,35 @@ sub editcomment ($$) {
 	$form->field(name => "type", value => $type, force => 1,
 		type => 'select', options => \@page_types);
 
-	$form->tmpl_param(username => $session->param('name'));
+	my $username=$session->param('name');
+	$form->tmpl_param(username => $username);
+		
+	$form->field(name => "subscribe", type => 'hidden');
+	$form->field(name => "anonsubscribe", type => 'hidden');
+	if (IkiWiki::Plugin::notifyemail->can("subscribe")) {
+		if (defined $username) {
+			$form->field(name => "subscribe", type => "checkbox",
+				options => [gettext("email replies to me")]);
+		}
+		elsif (IkiWiki::Plugin::passwordauth->can("anonuser")) {
+			$form->field(name => "anonsubscribe", type => "checkbox",
+				options => [gettext("email replies to me")]);
+		}
+	}
 
 	if ($config{comments_allowauthor} and
 	    ! defined $session->param('name')) {
 		$form->tmpl_param(allowauthor => 1);
 		$form->field(name => 'author', type => 'text', size => '40');
+		$form->field(name => 'email', type => 'text', size => '40');
 		$form->field(name => 'url', type => 'text', size => '40');
 	}
 	else {
 		$form->tmpl_param(allowauthor => 0);
 		$form->field(name => 'author', type => 'hidden', value => '',
 			force => 1);
+		$form->field(name => 'email', type => 'hidden', value => '',
+			force => 1);
 		$form->field(name => 'url', type => 'hidden', value => '',
 			force => 1);
 	}
@@ -393,6 +436,16 @@ sub editcomment ($$) {
 			$page));
 	}
 
+	# There's no UI to get here, but someone might construct the URL,
+	# leading to a comment that exists in the repository but isn't
+	# shown
+	if (!pagespec_match($page, $config{comments_pagespec},
+		location => $page)) {
+		error(sprintf(gettext(
+			"comments on page '%s' are not allowed"),
+			$page));
+	}
+
 	if (pagespec_match($page, $config{comments_closed_pagespec},
 		location => $page)) {
 		error(sprintf(gettext(
@@ -413,16 +466,16 @@ sub editcomment ($$) {
 		$username =~ s/"/&quot;/g;
 		$content .= " username=\"$username\"\n";
 	}
+
 	if (defined $session->param('nickname')) {
 		my $nickname = $session->param('nickname');
 		$nickname =~ s/"/&quot;/g;
 		$content .= " nickname=\"$nickname\"\n";
 	}
-	elsif (defined $session->remote_addr()) {
-		my $ip = $session->remote_addr();
-		if ($ip =~ m/^([.0-9]+)$/) {
-			$content .= " ip=\"$1\"\n";
-		}
+
+	if (!(defined $session->param('name') || defined $session->param('nickname')) &&
+		defined $session->remote_addr()) {
+		$content .= " ip=\"".$session->remote_addr()."\"\n";
 	}
 
 	if ($config{comments_allowauthor}) {
@@ -438,6 +491,12 @@ sub editcomment ($$) {
 		}
 	}
 
+	my $avatar=getavatar($session->param('name'));
+	if (defined $avatar && length $avatar) {
+		$avatar =~ s/"/&quot;/g;
+		$content .= " avatar=\"$avatar\"\n";
+	}
+
 	my $subject = $form->field('subject');
 	if (defined $subject && length $subject) {
 		$subject =~ s/"/&quot;/g;
@@ -446,8 +505,7 @@ sub editcomment ($$) {
 		$subject = "comment ".(num_comments($page, $config{srcdir}) + 1);
 	}
 	$content .= " subject=\"$subject\"\n";
-
-	$content .= " date=\"" . decode_utf8(strftime('%Y-%m-%dT%H:%M:%SZ', gmtime)) . "\"\n";
+	$content .= " date=\"" . commentdate() . "\"\n";
 
 	my $editcontent = $form->field('editcontent');
 	$editcontent="" if ! defined $editcontent;
@@ -478,6 +536,20 @@ sub editcomment ($$) {
 
 	if ($form->submitted eq POST_COMMENT && $form->validate) {
 		IkiWiki::checksessionexpiry($cgi, $session);
+
+		if (IkiWiki::Plugin::notifyemail->can("subscribe")) {
+			my $subspec="comment($page)";
+			if (defined $username &&
+			    length $form->field("subscribe")) {
+				IkiWiki::Plugin::notifyemail::subscribe(
+					$username, $subspec);
+			}
+			elsif (length $form->field("email") &&
+			       length $form->field("anonsubscribe")) {
+				IkiWiki::Plugin::notifyemail::anonsubscribe(
+					$form->field("email"), $subspec);
+			}
+		}
 		
 		$postcomment=1;
 		my $ok=IkiWiki::check_content(content => $form->field('editcontent'),
@@ -494,8 +566,8 @@ sub editcomment ($$) {
 		$postcomment=0;
 
 		if (! $ok) {
-			$location=unique_comment_location($page, $content, $config{srcdir}, "._comment_pending");
-			writefile("$location._comment_pending", $config{srcdir}, $content);
+			$location=unique_comment_location($page, $content, $IkiWiki::Plugin::transient::transientdir, "._comment_pending");
+			writefile("$location._comment_pending", $IkiWiki::Plugin::transient::transientdir, $content);
 
 			# Refresh so anything that deals with pending
 			# comments can be updated.
@@ -561,6 +633,36 @@ sub editcomment ($$) {
 	exit;
 }
 
+sub commentdate () {
+	strftime_utf8('%Y-%m-%dT%H:%M:%SZ', gmtime);
+}
+
+sub getavatar ($) {
+	my $user=shift;
+	return undef unless defined $user;
+
+	my $avatar;
+	eval q{use Libravatar::URL};
+	if (! $@) {
+		my $oiduser = eval { IkiWiki::openiduser($user) };
+		my $https=defined $config{url} && $config{url}=~/^https:/;
+
+		if (defined $oiduser) {
+			eval {
+				$avatar = libravatar_url(openid => $user, https => $https);
+			}
+		}
+		if (! defined $avatar &&
+		    (my $email = IkiWiki::userinfo_get($user, 'email'))) {
+			eval {
+				$avatar = libravatar_url(email => $email, https => $https);
+			}
+		}
+	}
+	return $avatar;
+}
+
+
 sub commentmoderation ($$) {
 	my $cgi=shift;
 	my $session=shift;
@@ -594,10 +696,17 @@ sub commentmoderation ($$) {
 				}
 
 				my $page=IkiWiki::dirname($f);
-				my $file="$config{srcdir}/$f";
+				my $filedir=$IkiWiki::Plugin::transient::transientdir;
+				my $file="$filedir/$f";
 				if (! -e $file) {
 					# old location
-					$file="$config{wikistatedir}/comments_pending/".$f;
+					$file="$config{srcdir}/$f";
+					$filedir=$config{srcdir};
+					if (! -e $file) {
+						# older location
+						$file="$config{wikistatedir}/comments_pending/".$f;
+						$filedir="$config{wikistatedir}/comments_pending";
+					}
 				}
 
 				if ($action eq 'Accept') {
@@ -612,7 +721,7 @@ sub commentmoderation ($$) {
 				}
 
 				require IkiWiki::Render;
-				IkiWiki::prune($file);
+				IkiWiki::prune($file, $filedir);
 			}
 		}
 
@@ -711,6 +820,8 @@ sub comments_pending () {
 		chdir($origdir) || die "chdir $origdir: $!";
 	};
 	
+	$find_comments->($IkiWiki::Plugin::transient::transientdir, "._comment_pending");
+	# old location
 	$find_comments->($config{srcdir}, "._comment_pending");
 	# old location
 	$find_comments->("$config{wikistatedir}/comments_pending/",
@@ -874,6 +985,11 @@ sub pagetemplate (@) {
 			$commentstate{$page}{commentauthorurl});
 	}
 
+	if ($template->query(name => 'commentauthoravatar')) {
+		$template->param(commentauthoravatar =>
+			$commentstate{$page}{commentauthoravatar});
+	}
+
 	if ($template->query(name => 'removeurl') &&
 	    IkiWiki::Plugin::remove->can("check_canremove") &&
 	    length $config{cgiurl}) {
@@ -897,7 +1013,7 @@ sub num_comments ($$) {
 	return int @comments;
 }
 
-sub unique_comment_location ($$$$) {
+sub unique_comment_location ($$$;$) {
 	my $page=shift;
 	eval q{use Digest::MD5 'md5_hex'};
 	error($@) if $@;
