@@ -8,6 +8,7 @@ use IkiWiki 3.00;
 
 sub import {
 	add_underlay("openid-selector");
+	add_underlay("jquery");
 	hook(type => "checkconfig", id => "openid", call => \&checkconfig);
 	hook(type => "getsetup", id => "openid", call => \&getsetup);
 	hook(type => "auth", id => "openid", call => \&auth);
@@ -77,14 +78,14 @@ sub openid_selector {
 
 	my $template=IkiWiki::template("openid-selector.tmpl");
 	$template->param(
-		cgiurl => $config{cgiurl},
+		cgiurl => IkiWiki::cgiurl(),
 		(defined $openid_error ? (openid_error => $openid_error) : ()),
 		(defined $openid_url ? (openid_url => $openid_url) : ()),
 		($real_cgi_signin ? (nonopenidform => $real_cgi_signin->($q, $session, 1)) : ()),
 	);
 
 	IkiWiki::printheader($session);
-	print IkiWiki::misctemplate("signin", $template->output);
+	print IkiWiki::cgitemplate($q, "signin", $template->output);
 	exit;
 }
 
@@ -99,10 +100,10 @@ sub formbuilder_setup (@) {
 	       IkiWiki::openiduser($session->param("name"))) {
 		$form->field(name => "openid_identifier", disabled => 1,
 			label => htmllink("", "", "ikiwiki/OpenID", noimageinline => 1),
-			value => $session->param("name"), 
-			size => length($session->param("name")), force => 1,
-			fieldset => "login");
-		$form->field(name => "email", type => "hidden");
+			value => "", 
+			size => 1, force => 1,
+			fieldset => "login",
+			comment => $session->param("name"));
 	}
 }
 
@@ -117,7 +118,9 @@ sub validate ($$$;$) {
 	my $claimed_identity = $csr->claimed_identity($openid_url);
 	if (! $claimed_identity) {
 		if ($errhandler) {
-			$errhandler->($csr->err);
+			if (ref($errhandler) eq 'CODE') {
+				$errhandler->($csr->err);
+			}
 			return 0;
 		}
 		else {
@@ -148,14 +151,14 @@ sub validate ($$$;$) {
 	}
 
 	my $cgiurl=$config{openid_cgiurl};
-	$cgiurl=$config{cgiurl} if ! defined $cgiurl;
+	$cgiurl=$q->url if ! defined $cgiurl;
 
 	my $trust_root=$config{openid_realm};
 	$trust_root=$cgiurl if ! defined $trust_root;
 
 	my $check_url = $claimed_identity->check_url(
-		return_to => "$cgiurl?do=postsignin",
-		trust_root => $trust_root,
+		return_to => auto_upgrade_https($q, "$cgiurl?do=postsignin"),
+		trust_root => auto_upgrade_https($q, $trust_root),
 		delayed_return => 1,
 	);
 	# Redirect the user to the OpenID server, which will
@@ -175,7 +178,7 @@ sub auth ($$) {
 			IkiWiki::redirect($q, $setup_url);
 		}
 		elsif ($csr->user_cancel) {
-			IkiWiki::redirect($q, $config{url});
+			IkiWiki::redirect($q, IkiWiki::baseurl(undef));
 		}
 		elsif (my $vident = $csr->verified_identity) {
 			$session->param(name => $vident->url);
@@ -187,12 +190,17 @@ sub auth ($$) {
 					$vident->signed_extension_fields('http://openid.net/srv/ax/1.0'),
 				);
 			}
+			my $nickname;
 			foreach my $ext (@extensions) {
 				foreach my $field (qw{value.email email}) {
 					if (exists $ext->{$field} &&
 					    defined $ext->{$field} &&
 					    length $ext->{$field}) {
 						$session->param(email => $ext->{$field});
+						if (! defined $nickname &&
+						    $ext->{$field}=~/(.+)@.+/) {
+							$nickname = $1;
+						}
 						last;
 					}
 				}
@@ -200,11 +208,15 @@ sub auth ($$) {
 					if (exists $ext->{$field} &&
 					    defined $ext->{$field} &&
 					    length $ext->{$field}) {
-						$session->param(username => $ext->{$field});
+						$nickname=$ext->{$field};
 						last;
 					}
 				}
 			}
+			if (defined $nickname) {
+				$session->param(nickname =>
+					Encode::decode_utf8($nickname));
+			}
 		}
 		else {
 			error("OpenID failure: ".$csr->err);
@@ -212,7 +224,7 @@ sub auth ($$) {
 	}
 	elsif (defined $q->param('openid_identifier')) {
 		# myopenid.com affiliate support
-		validate($q, $session, $q->param('openid_identifier'));
+		validate($q, $session, scalar $q->param('openid_identifier'));
 	}
 }
 
@@ -220,16 +232,17 @@ sub getobj ($$) {
 	my $q=shift;
 	my $session=shift;
 
+	eval q{use Net::INET6Glue::INET_is_INET6}; # may not be available
 	eval q{use Net::OpenID::Consumer};
 	error($@) if $@;
 
 	my $ua;
 	eval q{use LWPx::ParanoidAgent};
 	if (! $@) {
-		$ua=LWPx::ParanoidAgent->new;
+		$ua=LWPx::ParanoidAgent->new(agent => $config{useragent});
 	}
 	else {
-	        $ua=LWP::UserAgent->new;
+		$ua=useragent();
 	}
 
 	# Store the secret in the session.
@@ -240,16 +253,25 @@ sub getobj ($$) {
 	}
 	
 	my $cgiurl=$config{openid_cgiurl};
-	$cgiurl=$config{cgiurl} if ! defined $cgiurl;
+	$cgiurl=$q->url if ! defined $cgiurl;
 
 	return Net::OpenID::Consumer->new(
 		ua => $ua,
 		args => $q,
 		consumer_secret => sub { return shift()+$secret },
-		required_root => $cgiurl,
+		required_root => auto_upgrade_https($q, $cgiurl),
 	);
 }
 
+sub auto_upgrade_https {
+	my $q=shift;
+	my $url=shift;
+	if ($q->https()) {
+		$url=~s/^http:/https:/i;
+	}
+	return $url;
+}
+
 sub load_openid_module {
 	# Give up if module is unavailable to avoid needing to depend on it.
 	eval q{use Net::OpenID::Consumer};
