@@ -6,7 +6,7 @@ use warnings;
 use strict;
 use IkiWiki;
 
-my (%backlinks, %rendered);
+my (%backlinks, %rendered, %scanned);
 our %brokenlinks;
 my $links_calculated=0;
 
@@ -134,6 +134,7 @@ sub genpage ($$) {
 		ctime => displaytime($pagectime{$page}, undef, 1),
 		baseurl => baseurl($page),
 		html5 => $config{html5},
+		responsive_layout => $config{responsive_layout},
 	);
 
 	run_hooks(pagetemplate => sub {
@@ -154,6 +155,8 @@ sub genpage ($$) {
 
 sub scan ($) {
 	my $file=shift;
+	return if $phase > PHASE_SCAN || $scanned{$file};
+	$scanned{$file}=1;
 
 	debug(sprintf(gettext("scanning %s"), $file));
 
@@ -262,12 +265,13 @@ sub render ($$) {
 	}
 }
 
-sub prune ($) {
+sub prune ($;$) {
 	my $file=shift;
+	my $up_to=shift;
 
 	unlink($file);
 	my $dir=dirname($file);
-	while (rmdir($dir)) {
+	while ((! defined $up_to || $dir =~ m{^\Q$up_to\E\/}) && rmdir($dir)) {
 		$dir=dirname($dir);
 	}
 }
@@ -286,9 +290,22 @@ sub srcdir_check () {
 	
 }
 
-sub find_src_files () {
+# Finds all files in the srcdir, and the underlaydirs.
+# Returns the files, and their corresponding pages.
+#
+# When run in only_underlay mode, adds only the underlay files to
+# the files and pages passed in.
+sub find_src_files (;$$$) {
+	my $only_underlay=shift;
 	my @files;
+	if (defined $_[0]) {
+		@files=@{shift()};
+	}
 	my %pages;
+	if (defined $_[0]) {
+		%pages=%{shift()};
+	}
+
 	eval q{use File::Find};
 	error($@) if $@;
 
@@ -296,6 +313,8 @@ sub find_src_files () {
 	die $@ if $@;
 	my $origdir=getcwd();
 	my $abssrcdir=Cwd::abs_path($config{srcdir});
+	
+	@IkiWiki::underlayfiles=();
 
 	my ($page, $underlay);
 	my $helper=sub {
@@ -322,6 +341,7 @@ sub find_src_files () {
 			if (! -l "$abssrcdir/$f" && ! -e _) {
 				if (! $pages{$page}) {
 					push @files, $f;
+					push @IkiWiki::underlayfiles, $f;
 					$pages{$page}=1;
 				}
 			}
@@ -335,12 +355,14 @@ sub find_src_files () {
 		}
 	};
 
-	chdir($config{srcdir}) || die "chdir $config{srcdir}: $!";
-	find({
-		no_chdir => 1,
-		wanted => $helper,
-	}, '.');
-	chdir($origdir) || die "chdir $origdir: $!";
+	unless ($only_underlay) {
+		chdir($config{srcdir}) || die "chdir $config{srcdir}: $!";
+		find({
+			no_chdir => 1,
+			wanted => $helper,
+		}, '.');
+		chdir($origdir) || die "chdir $origdir: $!";
+	}
 
 	$underlay=1;
 	foreach (@{$config{underlaydirs}}, $config{underlaydir}) {
@@ -356,6 +378,50 @@ sub find_src_files () {
 	return \@files, \%pages;
 }
 
+# Given a hash of files that have changed, and a hash of files that were
+# deleted, should return the same results as find_src_files, with the same
+# sanity checks. But a lot faster!
+sub process_changed_files ($$) {
+	my $changed_raw=shift;
+	my $deleted_raw=shift;
+
+	my @files;
+	my %pages;
+
+	foreach my $file (keys %$changed_raw) {
+		my $page = pagename($file);
+		next if ! exists $pagesources{$page} && file_pruned($file);
+		my ($f) = $file =~ /$config{wiki_file_regexp}/; # untaint
+		if (! defined $f) {
+			warn(sprintf(gettext("skipping bad filename %s"), $file)."\n");
+			next;
+		}
+		push @files, $f;
+		if ($pages{$page}) {
+			debug(sprintf(gettext("%s has multiple possible source pages"), $page));
+		}
+		$pages{$page}=1;
+	}
+
+	# So far, we only have the changed files. Now add in all the old
+	# files that were not changed or deleted, excluding ones that came
+	# from the underlay.
+	my %old_underlay;
+	foreach my $f (@IkiWiki::underlayfiles) {
+		$old_underlay{$f}=1;
+	}
+	foreach my $page (keys %pagesources) {
+		my $f=$pagesources{$page};
+		unless ($old_underlay{$f} || exists $pages{$page} || exists $deleted_raw->{$f}) {
+			$pages{$page}=1;
+			push @files, $f;
+		}
+	}
+
+	# add in the underlay
+	find_src_files(1, \@files, \%pages);
+}
+
 sub find_new_files ($) {
 	my $files=shift;
 	my @new;
@@ -408,7 +474,8 @@ sub find_new_files ($) {
 			}
 			$pagecase{lc $page}=$page;
 			if (! exists $pagectime{$page}) {
-				$pagectime{$page}=(srcfile_stat($file))[10];
+				my $ctime=(srcfile_stat($file, 1))[10];
+				$pagectime{$page}=$ctime if defined $ctime;
 			}
 		}
 	}
@@ -447,7 +514,7 @@ sub remove_del (@) {
 		}
 	
 		foreach my $old (@{$oldrenderedfiles{$page}}) {
-			prune($config{destdir}."/".$old);
+			prune($config{destdir}."/".$old, $config{destdir});
 		}
 
 		foreach my $source (keys %destsources) {
@@ -468,10 +535,11 @@ sub find_changed ($) {
 	my @internal_changed;
 	foreach my $file (@$files) {
 		my $page=pagename($file);
-		my ($srcfile, @stat)=srcfile_stat($file);
-		if (! exists $pagemtime{$page} ||
-		    $stat[9] > $pagemtime{$page} ||
-	    	    $forcerebuild{$page}) {
+		my ($srcfile, @stat)=srcfile_stat($file, 1);
+		if (defined $srcfile && 
+		    (! exists $pagemtime{$page} ||
+		     $stat[9] > $pagemtime{$page} ||
+	    	     $forcerebuild{$page})) {
 			$pagemtime{$page}=$stat[9];
 
 			if (isinternal($page)) {
@@ -537,7 +605,7 @@ sub remove_unrendered () {
 		foreach my $file (@{$oldrenderedfiles{$page}}) {
 			if (! grep { $_ eq $file } @{$renderedfiles{$page}}) {
 				debug(sprintf(gettext("removing %s, no longer built by %s"), $file, $page));
-				prune($config{destdir}."/".$file);
+				prune($config{destdir}."/".$file, $config{destdir});
 			}
 		}
 	}
@@ -761,17 +829,42 @@ sub gen_autofile ($$$) {
 	return 1;
 }
 
-
 sub refresh () {
+	$phase = PHASE_SCAN;
+
 	srcdir_check();
 	run_hooks(refresh => sub { shift->() });
-	my ($files, $pages)=find_src_files();
-	my ($new, $internal_new)=find_new_files($files);
-	my ($del, $internal_del)=find_del_files($pages);
-	my ($changed, $internal_changed)=find_changed($files);
+	my ($files, $pages, $new, $internal_new, $del, $internal_del, $changed, $internal_changed);
+	my $want_find_changes=$config{only_committed_changes} &&
+		exists $IkiWiki::hooks{rcs}{rcs_find_changes} &&
+		exists $IkiWiki::hooks{rcs}{rcs_get_current_rev};
+	if (! $config{rebuild} && $want_find_changes && defined $IkiWiki::lastrev && length $IkiWiki::lastrev) {
+		my ($changed_raw, $del_raw);
+		($changed_raw, $del_raw, $IkiWiki::lastrev) = $IkiWiki::hooks{rcs}{rcs_find_changes}{call}->($IkiWiki::lastrev);
+		($files, $pages)=process_changed_files($changed_raw, $del_raw);
+	}
+	else {
+		($files, $pages)=find_src_files();
+	}
+	if ($want_find_changes) {
+		if (! defined($IkiWiki::lastrev) || ! length $IkiWiki::lastrev) {
+			$IkiWiki::lastrev=$IkiWiki::hooks{rcs}{rcs_get_current_rev}{call}->();
+		}
+	}
+	($new, $internal_new)=find_new_files($files);
+	($del, $internal_del)=find_del_files($pages);
+	($changed, $internal_changed)=find_changed($files);
+	my %existingfiles;
 	run_hooks(needsbuild => sub {
 		my $ret=shift->($changed, [@$del, @$internal_del]);
-		$changed=$ret if ref $ret eq 'ARRAY';
+		if (ref $ret eq 'ARRAY' && $ret != $changed) {
+			if (! %existingfiles) {
+				foreach my $f (@$files) {
+					$existingfiles{$f}=1;
+				}
+			}
+			@$changed=grep $existingfiles{$_}, @$ret;
+		}
 	});
 	my $oldlink_targets=calculate_old_links($changed, $del);
 
@@ -790,7 +883,13 @@ sub refresh () {
 	}
 
 	calculate_links();
-	
+
+	# At this point it becomes OK to start matching pagespecs.
+	$phase = PHASE_RENDER;
+	# Save some memory: we no longer need to keep track of which pages
+	# we've scanned
+	%scanned = ();
+
 	remove_del(@$del, @$internal_del);
 
 	foreach my $file (@$changed) {
@@ -829,8 +928,13 @@ sub refresh () {
 		run_hooks(delete => sub { shift->(@$del, @$internal_del) });
 	}
 	if (%rendered) {
-		run_hooks(change => sub { shift->(keys %rendered) });
+		run_hooks(rendered => sub { shift->(keys %rendered) });
+		run_hooks(change => sub { shift->(keys %rendered) }); # back-compat
 	}
+	my %all_changed = map { $_ => 1 }
+		@$new, @$changed, @$del,
+		@$internal_new, @$internal_changed, @$internal_del;
+	run_hooks(changes => sub { shift->(keys %all_changed) });
 }
 
 sub clean_rendered {
@@ -839,7 +943,7 @@ sub clean_rendered {
 	remove_unrendered();
 	foreach my $page (keys %oldrenderedfiles) {
 		foreach my $file (@{$oldrenderedfiles{$page}}) {
-			prune($config{destdir}."/".$file);
+			prune($config{destdir}."/".$file, $config{destdir});
 		}
 	}
 }
@@ -849,6 +953,10 @@ sub commandline_render () {
 	loadindex();
 	unlockwiki();
 
+	# This function behaves as though it's in the render phase;
+	# all other files are assumed to have been scanned last time.
+	$phase = PHASE_RENDER;
+
 	my $srcfile=possibly_foolish_untaint($config{render});
 	my $file=$srcfile;
 	$file=~s/\Q$config{srcdir}\E\/?//;
